#!/usr/bin/env node
"use strict";
var _a, _b;
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const swagger_parser_1 = tslib_1.__importDefault(require("@apidevtools/swagger-parser"));
const ts_morph_1 = require("ts-morph");
const path_1 = tslib_1.__importDefault(require("path"));
const templateApi_1 = require("./templateApi");
const fs = tslib_1.__importStar(require("fs"));
const serializefunc_1 = require("./serializefunc");
const utils_1 = require("./utils");
const yargs_1 = tslib_1.__importDefault(require("yargs"));
const runtime_1 = require("./runtime");
const folder_hash_1 = require("folder-hash");
const version_json_1 = require("./version.json");
const pluralize_1 = require("pluralize");
const argv = yargs_1.default.command('generate', 'generate a TypeScript client').demandCommand().string('output').describe({ 'output': 'the output directory' }).demandOption('output').parseSync();
const output = (_a = argv['output']) !== null && _a !== void 0 ? _a : path_1.default.join(process.cwd(), 'build');
const file = (_b = argv._[0]) === null || _b === void 0 ? void 0 : _b.toString();
if (!file && argv.$0 !== 'generate') {
    yargs_1.default.showHelp();
    process.exit(1);
}
fs.mkdirSync(output, { recursive: true });
const hashFile = path_1.default.join(output, '.hash');
const project = new ts_morph_1.Project({});
(() => tslib_1.__awaiter(void 0, void 0, void 0, function* () {
    var _c, _d;
    const hash = yield (0, folder_hash_1.hashElement)(path_1.default.dirname(file), { algo: 'sha256' });
    const versionizedHash = `${version_json_1.version}::${hash.hash}`;
    if (fs.existsSync(hashFile)) {
        const oldHash = fs.readFileSync(hashFile, { encoding: 'utf8' });
        if (versionizedHash === oldHash) {
            console.log("Client is up-to-date:", versionizedHash);
            process.exit(0);
        }
    }
    console.log('Parsing', file);
    const spec = yield swagger_parser_1.default.parse(file);
    console.log('Resolving', file);
    const resolver = yield swagger_parser_1.default.resolve(file);
    const modelsFile = path_1.default.join(output, 'models.ts');
    console.log('Creating', modelsFile);
    const models = project.createSourceFile(modelsFile, '/* tslint:disable */\n/* eslint-disable */', { overwrite: true });
    const ctx = { file: models, resolver: resolver, refPath: [], currFile: path_1.default.basename(file) };
    Object.entries((_d = (_c = spec.components) === null || _c === void 0 ? void 0 : _c.schemas) !== null && _d !== void 0 ? _d : {}).forEach(([key, value]) => getAndCreateType(ctx, key, value));
    const pathCtx = { file: models, resolver: resolver, refPath: [], currFile: path_1.default.basename(file) };
    const apis = {};
    Object.entries(spec.paths).forEach(([path, body]) => {
        handleObjectOperation(pathCtx, apis, 'GET', body === null || body === void 0 ? void 0 : body.get, path);
        handleObjectOperation(pathCtx, apis, 'POST', body === null || body === void 0 ? void 0 : body.post, path);
        handleObjectOperation(pathCtx, apis, 'PUT', body === null || body === void 0 ? void 0 : body.put, path);
        handleObjectOperation(pathCtx, apis, 'DELETE', body === null || body === void 0 ? void 0 : body.delete, path);
        handleObjectOperation(pathCtx, apis, 'PATCH', body === null || body === void 0 ? void 0 : body.patch, path);
    });
    Object.values(apis).forEach((bundle) => {
        const apiFile = path_1.default.join(output, bundle.name + 'Api.ts');
        console.log('Creating', apiFile);
        project.createSourceFile(apiFile, (0, templateApi_1.createApi)(models, bundle), { overwrite: true }).formatText();
    });
    const toExportStr = (name, suffix = '') => `export * from './${name}${suffix}';`;
    const generalExports = ['models', 'runtime'].map(name => toExportStr(name));
    const apiExports = Object.keys(apis).map(api => toExportStr(api, 'Api'));
    console.log('Creating (de)serialize functions');
    (0, serializefunc_1.createDeSerializeFunctions)(models);
    models.formatText();
    const runtimeFile = path_1.default.join(output, 'runtime.ts');
    console.log('Creating', runtimeFile);
    project.createSourceFile(runtimeFile, runtime_1.runtimeContent, { overwrite: true }).formatText();
    const indexFile = path_1.default.join(output, 'index.ts');
    console.log('Creating', indexFile);
    project.createSourceFile(indexFile, [...generalExports, ...apiExports].join('\n'), { overwrite: true }).formatText();
    project.saveSync();
    console.log('Creating', hashFile);
    fs.writeFileSync(hashFile, versionizedHash, { encoding: 'utf8' });
}))();
const handleObjectOperation = (ctx, apis, method, op, path) => {
    var _a, _b, _c, _d, _e, _f;
    if (!op) {
        return;
    }
    const name = (_b = (_a = op.tags) === null || _a === void 0 ? void 0 : _a[0]) !== null && _b !== void 0 ? _b : 'unset';
    apis[name] = (_c = apis[name]) !== null && _c !== void 0 ? _c : { name, calls: [] };
    if (!op.operationId) {
        throw new Error("operation id not set");
    }
    const call = {
        method,
        nickname: op.operationId,
        path,
        body: op.requestBody && getAndCreateRequestBody(ctx, name, op.requestBody),
        returnType: getAndCreateResponseBody(ctx, name, (_d = op.responses) === null || _d === void 0 ? void 0 : _d[200]),
        params: (_f = (_e = op.parameters) === null || _e === void 0 ? void 0 : _e.map((param) => getAndCreateParameter(ctx, name, param)).filter((o) => !!o)) !== null && _f !== void 0 ? _f : [],
    };
    apis[name].calls.push(call);
};
const getAndCreateParameter = (ctx, name, obj) => {
    var _a;
    if (!obj) {
        return undefined;
    }
    if ('$ref' in obj) {
        throw new Error('not supported');
    }
    return {
        in: obj.in,
        required: (_a = obj.required) !== null && _a !== void 0 ? _a : false,
        type: getAndCreateType(ctx, name, obj.schema),
        interfaceName: adjustParameterName(obj.name, obj.schema),
        name: obj.name
    };
};
const getAndCreateRequestBody = (ctx, name, obj) => {
    if ('$ref' in obj) {
        return (0, utils_1.resolveReference)(ctx, obj.$ref, ({ newCtx, name, resolved }) => getAndCreateRequestBody(newCtx, name, resolved));
    }
    const json = obj.content['application/json'];
    if (json.schema) {
        return getAndCreateType(ctx, name, json.schema);
    }
    return 'unknown';
};
const getAndCreateResponseBody = (ctx, name, obj) => {
    var _a;
    if (!obj) {
        return undefined;
    }
    if ('$ref' in obj) {
        return (0, utils_1.resolveReference)(ctx, obj.$ref, ({ newCtx, name, resolved }) => getAndCreateResponseBody(newCtx, name, resolved));
    }
    const json = (_a = obj.content) === null || _a === void 0 ? void 0 : _a['application/json'];
    if (json === null || json === void 0 ? void 0 : json.schema) {
        return getAndCreateType(ctx, name, json.schema);
    }
    return undefined;
};
const getAndCreateType = (ctx, name, obj) => {
    var _a, _b, _c, _d, _e;
    if ('$ref' in obj) {
        return (0, utils_1.resolveReference)(ctx, obj.$ref, ({ newCtx, name, resolved }) => getAndCreateType(newCtx, name, resolved));
    }
    if (obj.type === 'string') {
        if (obj.format === 'date-time') {
            return 'Date';
        }
        const enumValues = (_b = (_a = obj.enum) !== null && _a !== void 0 ? _a : obj["x-enum-varnames"]) !== null && _b !== void 0 ? _b : false;
        if (enumValues === false) {
            return 'string';
        }
        if (!ctx.file.getEnum(name)) {
            ctx.file.addEnum({
                isExported: true,
                name: name,
                members: enumValues.map((value) => {
                    return ({ name: (0, utils_1.pascalCase)(value), value: value });
                })
            });
        }
        return name;
    }
    if (obj.type === 'number' || obj.type === 'integer') {
        return 'number';
    }
    if (obj.type === 'boolean') {
        return 'boolean';
    }
    if (obj.type === 'array') {
        return getAndCreateType(ctx, name, obj.items) + '[]';
    }
    if (obj.type === 'object') {
        if (obj.title) {
            name = obj.title;
        }
        if (obj.additionalProperties && typeof obj.additionalProperties !== 'boolean') {
            if (ctx.file.getInterface(name)) {
                return name;
            }
            ctx.file.addInterface({
                name,
                isExported: true,
                indexSignatures: [
                    { keyName: 'key', keyType: 'string', returnType: getAndCreateType(ctx, name + "Inner", obj.additionalProperties) },
                ],
            });
            return name;
        }
        if (!obj.properties) {
            return 'unknown';
        }
        if (name.endsWith(utils_1.UnionBaseSuffix)) {
            console.log(`names shouldn't end with Base, this is reserved for union types`);
        }
        const useName = name;
        if (obj.discriminator) {
            name = name + utils_1.UnionBaseSuffix;
        }
        if (ctx.file.getInterface(name)) {
            return useName;
        }
        ctx.file.addInterface({
            isExported: true,
            name,
        }).addProperties(Object.entries((_c = obj.properties) !== null && _c !== void 0 ? _c : {}).map(([propName, body]) => {
            var _a;
            const type = getAndCreateType(ctx, name + (0, utils_1.pascalCase)(propName), body);
            return {
                hasQuestionToken: !((_a = obj.required) === null || _a === void 0 ? void 0 : _a.includes(propName)),
                name: propName,
                type,
            };
        }));
        return useName;
    }
    if (obj.allOf) {
        if (ctx.file.getInterface(name)) {
            return name;
        }
        if (obj.allOf.length !== 2 && obj.allOf.length !== 1) {
            throw new Error('allOf may only have one or twe entries ' + name);
        }
        const [base, data] = obj.allOf;
        const baseType = getAndCreateType(ctx, name + 'AllOfChild', base);
        if (obj.allOf.length === 2) {
            if (!('type' in data) || data.type !== 'object') {
                throw new Error('allOf second parameter must be type object and may never be a reference ' + name);
            }
            const newInterface = ctx.file.addInterface({
                isExported: true,
                name,
                extends: [baseType + utils_1.UnionBaseSuffix]
            });
            newInterface.addProperties(Object.entries((_d = data.properties) !== null && _d !== void 0 ? _d : {}).map(([propName, body]) => {
                var _a;
                const type = getAndCreateType(ctx, name + (0, utils_1.pascalCase)(propName), body);
                return {
                    hasQuestionToken: !((_a = data.required) === null || _a === void 0 ? void 0 : _a.includes(propName)),
                    name: propName,
                    type,
                };
            }));
            newInterface.addProperty({ name: '$type', type: `'${name}'` });
        }
        else {
            const newElement = {
                isExported: true,
                name,
                extends: [baseType + utils_1.UnionBaseSuffix],
                properties: [
                    { name: '$type', type: `'${name}'` },
                ],
            };
            ctx.file.addInterface(newElement);
        }
        const existing = ctx.file.getTypeAlias(baseType);
        if (existing) {
            existing.setType(((_e = existing.getTypeNode()) === null || _e === void 0 ? void 0 : _e.getText()) + ' | ' + name);
        }
        else {
            ctx.file.addTypeAlias({
                isExported: true,
                name: baseType,
                type: name,
            });
        }
        return name;
    }
    console.error('unhandled type', obj);
    process.exit(1);
};
const adjustParameterName = (name, obj) => {
    if ('$ref' in obj) {
        return name;
    }
    if (obj.type === 'array') {
        return (0, pluralize_1.plural)(name);
    }
    return name;
};
//# sourceMappingURL=index.js.map